@@ -10,26 +10,26 @@ import (
 	"io"
 	"os"
 	"os/exec"
-	"reflect"
 	"regexp"
 	"strings"
 	"sync"
 	"testing"
 
-	"github.com/kr/pretty"
+	"github.com/google/go-cmp/cmp"
 )
 
-func TestKeepComments(t *testing.T) {
+func TestParseBashKeepComments(t *testing.T) {
 	t.Parallel()
-	in := "# foo\ncmd\n# bar"
-	want := &File{
-		Stmts: []*Stmt{{
-			Comments: []Comment{{Text: " foo"}},
-			Cmd:      litCall("cmd"),
-		}},
-		Last: []Comment{{Text: " bar"}},
+	p := NewParser(KeepComments(true))
+	for i, c := range fileTestsKeepComments {
+		want := c.Bash
+		if want == nil {
+			continue
+		}
+		for j, in := range c.Strs {
+			t.Run(fmt.Sprintf("#%03d-%d", i, j), singleParse(p, in, want))
+		}
 	}
-	singleParse(NewParser(KeepComments(true)), in, want)(t)
 }
 
 func TestParseBash(t *testing.T) {
@@ -155,17 +155,10 @@ func TestParseBats(t *testing.T) {
 }
 
 func TestMain(m *testing.M) {
-	// Set the locale to computer-friendly English and UTF-8.
-	// Some systems like Arch miss C.UTF8, so fall back to the US English locale.
-	if out, _ := exec.Command("locale", "-a").Output(); strings.Contains(
-		strings.ToLower(string(out)), "c.utf",
-	) {
-		os.Setenv("LANGUAGE", "C.UTF-8")
-		os.Setenv("LC_ALL", "C.UTF-8")
-	} else {
-		os.Setenv("LANGUAGE", "en_US.UTF-8")
-		os.Setenv("LC_ALL", "en_US.UTF-8")
-	}
+	// Set the locale to computer-friendly English and UTF-8, C.UTF-8,
+	// which started shipping with glibc 2.35 in February 2022.
+	os.Setenv("LANGUAGE", "C.UTF-8")
+	os.Setenv("LC_ALL", "C.UTF-8")
 	os.Exit(m.Run())
 }
 
@@ -224,8 +217,12 @@ var extGlobRe = regexp.MustCompile(`[@?*+!]\(`)
 
 func confirmParse(in, cmd string, wantErr bool) func(*testing.T) {
 	return func(t *testing.T) {
+		t.Helper()
 		t.Parallel()
 		var opts []string
+		if strings.Contains(in, "\\\r\n") {
+			t.Skip("shells do not generally support CRLF line endings")
+		}
 		if cmd == "bash" && extGlobRe.MatchString(in) {
 			// otherwise bash refuses to parse these
 			// properly. Also avoid -n since that too makes
@@ -235,24 +232,22 @@ func confirmParse(in, cmd string, wantErr bool) func(*testing.T) {
 			// -n makes bash accept invalid inputs like
 			// "let" or "`{`", so only use it in
 			// non-erroring tests. Should be safe to not use
-			// -n anyway since these are supposed to just
-			// fail.
+			// -n anyway since these are supposed to just fail.
 			// also, -n will break if we are using extglob
 			// as extglob is not actually applied.
 			opts = append(opts, "-n")
 		}
 		cmd := exec.Command(cmd, opts...)
+		cmd.Dir = t.TempDir() // to be safe
 		cmd.Stdin = strings.NewReader(in)
 		var stderr bytes.Buffer
 		cmd.Stderr = &stderr
 		err := cmd.Run()
 		if stderr.Len() > 0 {
 			// bash sometimes likes to error on an input via stderr
-			// while forgetting to set the exit code to non-zero.
-			// Fun.
-			if s := stderr.String(); !strings.Contains(s, ": warning: ") {
-				err = errors.New(s)
-			}
+			// while forgetting to set the exit code to non-zero. Fun.
+			// Note that we also treat warnings as errors.
+			err = errors.New(stderr.String())
 		}
 		if err != nil && strings.Contains(err.Error(), "command not found") {
 			err = nil
@@ -379,6 +374,8 @@ func TestParseErrMirBSDKornConfirm(t *testing.T) {
 	}
 }
 
+var cmpOpt = cmp.FilterValues(func(p1, p2 Pos) bool { return true }, cmp.Ignore())
+
 func singleParse(p *Parser, in string, want *File) func(t *testing.T) {
 	return func(t *testing.T) {
 		t.Helper()
@@ -386,10 +383,9 @@ func singleParse(p *Parser, in string, want *File) func(t *testing.T) {
 		if err != nil {
 			t.Fatalf("Unexpected error in %q: %v", in, err)
 		}
-		clearPosRecurse(t, in, got)
-		if !reflect.DeepEqual(got, want) {
-			t.Fatalf("syntax tree mismatch in %q\ndiff:\n%s", in,
-				strings.Join(pretty.Diff(want, got), "\n"))
+		recursiveSanityCheck(t, in, got)
+		if diff := cmp.Diff(want, got, cmpOpt); diff != "" {
+			t.Errorf("syntax tree mismatch in %q (-want +got):\n%s", in, diff)
 		}
 	}
 }
@@ -423,9 +419,9 @@ func BenchmarkParse(b *testing.B) {
 
 type errorCase struct {
 	in          string
-	common      interface{}
-	bash, posix interface{}
-	bsmk, mksh  interface{}
+	common      any
+	bash, posix any
+	bsmk, mksh  any
 }
 
 var shellTests = []errorCase{
@@ -805,6 +801,14 @@ var shellTests = []errorCase{
 		in:     "echo && > #",
 		common: `1:9: > must be followed by a word`,
 	},
+	{
+		in:    "foo &>/dev/null",
+		posix: `1:5: &> redirects are a bash/mksh feature`,
+	},
+	{
+		in:    "foo &>>/dev/null",
+		posix: `1:5: &> redirects are a bash/mksh feature`,
+	},
 	{
 		in:     "<<",
 		common: `1:1: << must be followed by a word`,
@@ -812,17 +816,17 @@ var shellTests = []errorCase{
 	{
 		in:     "<<EOF",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<EOF\n\\",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<EOF\n\\\n",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in: "<<EOF\n\\\nEOF",
@@ -832,52 +836,59 @@ var shellTests = []errorCase{
 	{
 		in:     "<<EOF\nfoo\\\nEOF",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<'EOF'\n\\\n",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<EOF <`\n#\n`\n``",
 		common: `1:1: unclosed here-document 'EOF'`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<'EOF'",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<\\EOF",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<\\\\EOF",
 		common: `1:1: unclosed here-document '\EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document '\EOF'`,
+		bsmk:   `1:1: unclosed here-document '\EOF'`,
 	},
 	{
 		in:     "<<-EOF",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<-EOF\n\t",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<-'EOF'\n\t",
 		common: `1:1: unclosed here-document 'EOF' #NOERR`,
-		mksh:   `1:1: unclosed here-document 'EOF'`,
+		bsmk:   `1:1: unclosed here-document 'EOF'`,
 	},
 	{
 		in:     "<<\nEOF\nbar\nEOF",
 		common: `1:1: << must be followed by a word`,
 	},
+	{
+		in:   "$(<<EOF\nNOTEOF)",
+		bsmk: `1:3: unclosed here-document 'EOF'`,
+	},
+	{
+		in:   "`<<EOF\nNOTEOF`",
+		bsmk: `1:2: unclosed here-document 'EOF'`,
+	},
 	{
 		in:     "if",
 		common: `1:1: "if" must be followed by a statement list`,
@@ -1316,7 +1327,7 @@ var shellTests = []errorCase{
 	},
 	{
 		in:     "`\"`",
-		common: "1:3: reached EOF without closing quote `",
+		common: "1:2: reached ` without closing quote \"",
 	},
 	{
 		in:     "`\\```",
@@ -1339,11 +1350,19 @@ var shellTests = []errorCase{
 		in:     "<<${bar}\n${bar}",
 		common: `1:3: expansions not allowed in heredoc words #NOERR`,
 	},
+
+	// bash uses "$(bar)" as the closing word, but other shells use "$".
+	// We instead give an error for expansions in heredoc words.
 	{
 		in:    "<<$(bar)\n$",
-		bsmk:  `1:3: expansions not allowed in heredoc words #NOERR`,
 		posix: `1:3: expansions not allowed in heredoc words`,
+		mksh:  `1:3: expansions not allowed in heredoc words #NOERR`,
 	},
+	{
+		in:   "<<$(bar)\n$(bar)",
+		bash: `1:3: expansions not allowed in heredoc words #NOERR`,
+	},
+
 	{
 		in:     "<<$-\n$-",
 		common: `1:3: expansions not allowed in heredoc words #NOERR`,
@@ -1837,7 +1856,15 @@ var shellTests = []errorCase{
 	},
 	{
 		in:    "function foo() { bar; }",
-		posix: `1:13: a command can only contain words and redirects; encountered (`,
+		posix: `1:13: the "function" builtin exists in bash; tried parsing as posix`,
+	},
+	{
+		in:    "declare foo=(bar)",
+		posix: `1:13: the "declare" builtin exists in bash; tried parsing as posix`,
+	},
+	{
+		in:    "let foo=(bar)",
+		posix: `1:9: the "let" builtin exists in bash; tried parsing as posix`,
 	},
 	{
 		in:    "echo <(",
@@ -1899,7 +1926,21 @@ var shellTests = []errorCase{
 	},
 	{
 		in:    "echo ${!foo}",
-		posix: `1:8: ${!foo} is a bash/mksh feature`,
+		posix: `1:6: "${!foo}" is a bash/mksh feature`,
+	},
+	{
+		in:    "echo ${!foo*}",
+		posix: `1:6: "${!foo*}" is a bash feature`,
+		mksh:  `1:6: "${!foo*}" is a bash feature`,
+	},
+	{
+		in:    "echo ${!foo@}",
+		posix: `1:12: this expansion operator is a bash/mksh feature`,
+		mksh:  `1:6: "${!foo@}" is a bash feature`,
+	},
+	{
+		in:    "echo ${!foo[@]}",
+		posix: `1:12: arrays are a bash/mksh feature`,
 	},
 	{
 		in:    "echo ${foo[1]}",
@@ -1913,6 +1954,14 @@ var shellTests = []errorCase{
 		in:    "echo ${foo:1}",
 		posix: `1:11: slicing is a bash/mksh feature`,
 	},
+	{
+		in:    "foo <<< bar",
+		posix: `1:5: herestrings are a bash/mksh feature`,
+	},
+	{
+		in:    "foo << < bar",
+		posix: `1:5: << must be followed by a word`,
+	},
 	{
 		in:    "echo ${foo,bar}",
 		posix: `1:11: this expansion operator is a bash feature`,
@@ -1960,7 +2009,7 @@ var shellTests = []errorCase{
 	},
 	{
 		in:     "`\"`\\",
-		common: "1:3: reached EOF without closing quote `",
+		common: "1:2: reached ` without closing quote \"",
 	},
 }
 
@@ -2248,11 +2297,10 @@ func TestParseDocument(t *testing.T) {
 			if err != nil {
 				t.Fatal(err)
 			}
-			clearPosRecurse(t, "", got)
+			recursiveSanityCheck(t, "", got)
 			want := &Word{Parts: tc.want}
-			if !reflect.DeepEqual(got, want) {
-				t.Fatalf("syntax tree mismatch in %q\ndiff:\n%s", tc.in,
-					strings.Join(pretty.Diff(want, got), "\n"))
+			if diff := cmp.Diff(want, got, cmpOpt); diff != "" {
+				t.Errorf("syntax tree mismatch in %q (-want +got):\n%s", tc.in, diff)
 			}
 		})
 	}
@@ -2345,10 +2393,9 @@ func TestParseArithmetic(t *testing.T) {
 			if err != nil {
 				t.Fatal(err)
 			}
-			clearPosRecurse(t, "", got)
-			if !reflect.DeepEqual(got, tc.want) {
-				t.Fatalf("syntax tree mismatch in %q\ndiff:\n%s", tc.in,
-					strings.Join(pretty.Diff(tc.want, got), "\n"))
+			recursiveSanityCheck(t, "", got)
+			if diff := cmp.Diff(tc.want, got, cmpOpt); diff != "" {
+				t.Errorf("syntax tree mismatch in %q (-want +got):\n%s", tc.in, diff)
 			}
 		})
 	}
@@ -2369,7 +2416,7 @@ func TestParseArithmeticError(t *testing.T) {
 var stopAtTests = []struct {
 	in   string
 	stop string
-	want interface{}
+	want any
 }{
 	{
 		"foo bar", "$$",
